/*
  +----------------------------------------------------------------------+
  | PHP Version 4                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2003 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 2.02 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available at through the world-wide-web at                           |
  | http://www.php.net/license/2_02.txt.                                 |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: John McCaskey (jmccaskey@gmail.com)                          |
  +----------------------------------------------------------------------+

  $Id: header,v 1.10.8.1 2003/07/14 15:59:18 sniper Exp $ 
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "ext/standard/php_string.h"
#include "ext/standard/php_var.h"
#include "ext/standard/php_smart_str.h"
#include "php_mcache.h"
#include "memcache.h"

/* True global resources - no need for thread safety here */
static int le_mcache, le_pmcache;
static zend_class_entry *mcache_class_entry_ptr;
static struct memcache_ctxt *persist_ctx;
static struct memcache_ctxt *standard_ctx;

/* {{{ Memory allocation wrappers
 * Memory allocation wrapper functions, because we can't pass the address
 * of the define's to libmemcache... 
 */
void *xemalloc(size_t size) {
	return _emalloc((size) ZEND_FILE_LINE_CC ZEND_FILE_LINE_EMPTY_CC);
}

void xefree(void *ptr) {
	return _efree((ptr) ZEND_FILE_LINE_CC ZEND_FILE_LINE_EMPTY_CC);
}

void *xerealloc(void *ptr, size_t size) {
	return _erealloc((ptr), (size), 0 ZEND_FILE_LINE_CC ZEND_FILE_LINE_EMPTY_CC);
}

void *xpemalloc(size_t size) {
	return pemalloc(size, 1);
}

void *xperealloc(void *ptr, size_t size) {
	return perealloc(ptr, size, 1);
}

void xpefree(void *ptr) {
	return pefree(ptr, 1);
}

/* }}} */

/* {{{ Error handling callback function.
 * libmemcache will call this function for us after we setup the initial context.
 */
void mcache_err(MCM_ERR_FUNC_ARGS) {
	const struct memcache_ctxt *ctxt;
	struct memcache_err_ctxt *ectxt;
  const char *errno_str, *severity;
  
	int e_level = E_NOTICE;

	MCM_ERR_INIT_CTXT(ctxt, ectxt);

	if(ectxt->severity <= MCM_ERR_LVL_NOTICE) {
		e_level = E_NOTICE;
	} else if(ectxt->severity <= MCM_ERR_LVL_WARN) {
		e_level = E_WARNING;
	} else {
		e_level = E_ERROR;
	}

  if (ectxt->errnum != 0)
    errno_str = strerror(ectxt->errnum);
  else
    errno_str = NULL;
  
  if (ectxt->errmsg != NULL && errno_str != NULL && ectxt->errmsg != NULL)
    zend_error(e_level, "%s():%u: %s: %s: %.*s\n", ectxt->funcname, ectxt->lineno, ectxt->errstr, errno_str, (int)ectxt->errlen, ectxt->errmsg);
  else if (ectxt->errmsg == NULL && errno_str != NULL && ectxt->errmsg != NULL)
    zend_error(e_level, "%s():%u: %s: %.*s\n", ectxt->funcname, ectxt->lineno, errno_str, (int)ectxt->errlen, ectxt->errmsg);
  else if (ectxt->errmsg != NULL && errno_str == NULL && ectxt->errmsg != NULL)
    zend_error(e_level, "%s():%u: %s: %.*s\n", ectxt->funcname, ectxt->lineno, ectxt->errstr, (int)ectxt->errlen, ectxt->errmsg);
  else if (ectxt->errmsg != NULL && errno_str != NULL && ectxt->errmsg == NULL)
    zend_error(e_level, "%s():%u: %s: %s\n", ectxt->funcname, ectxt->lineno, errno_str, ectxt->errstr);
  else if (ectxt->errmsg == NULL && errno_str == NULL && ectxt->errmsg != NULL)
    zend_error(e_level, "%s():%u: %.*s\n", ectxt->funcname, ectxt->lineno, (int)ectxt->errlen, ectxt->errmsg);
  else if (ectxt->errmsg == NULL && errno_str != NULL && ectxt->errmsg == NULL)
    zend_error(e_level, "%s():%u: %s\n", ectxt->funcname, ectxt->lineno, errno_str);
  else if (ectxt->errmsg != NULL && errno_str == NULL && ectxt->errmsg == NULL)
    zend_error(e_level, "%s():%u: %s\n", ectxt->funcname, ectxt->lineno, ectxt->errmsg);
  else
    zend_error(e_level, "%s():%u\n", ectxt->funcname, ectxt->lineno);

	/**
	 * An error getting data from memcache should never
	 * cause the users apache process to exit/abort!
	 * 
	 * As of libmemcache 1.3.0.beta2 we could optimize
	 * by only setting this for MCM_ERR_LVL_WARN or greater,
	 * but to be safe for possible future changes I just
	 * always set it.
	 */
	ectxt->cont = 'y';
}


/* {{{ mcache_functions[]
 *
 * Every user visible function must have an entry in mcache_functions[].
 */
function_entry mcache_functions[] = {
	PHP_FE(memcache,	NULL)		/* Factory method */
	PHP_FE(pmemcache,	NULL) 		/* Factory method for persistent objects */
	{NULL, NULL, NULL}	/* Must be the last line in mcache_functions[] */
};

static zend_function_entry mcache_class_functions[] = {
	PHP_FE(is_initialized, NULL) 	/* Tells whether this is a brand new memcache object or not, use to decide whether to add servers. */
	PHP_FE(set_initialized, NULL)   /* Set the initialized flag on the memcache object. */
	PHP_FE(add_server, NULL)		/* Add a server to the memcache object. */
	PHP_FE(get, NULL)	/* Get item out of cache. */
	PHP_FE(set, NULL)	/* Set item in the cache. */
	PHP_FE(add, NULL)	/* Add an item to cache (only if it doesn't exist) */
	PHP_FE(replace, NULL)	/* Replace an item in the cache (only if it does exist) */
	PHP_FE(incr, NULL)	/* Increment key. */
	PHP_FE(decr, NULL) /* Decrement key. */
	PHP_FE(delete, NULL)	/* Delete key from cache. */
	PHP_FE(flush_all, NULL) /* Flush all keys from all servers */
	PHP_FE(stats, NULL) /* Get statistics from the memcache server */
	{NULL, NULL, NULL} /* Sentinel */
};

/* }}} */

/* {{{ mcache_module_entry
 */
zend_module_entry mcache_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"mcache",
	mcache_functions,
	PHP_MINIT(mcache),
	PHP_MSHUTDOWN(mcache),
	PHP_RINIT(mcache),		/* Replace with NULL if there's nothing to do at request start */
	PHP_RSHUTDOWN(mcache),	/* Replace with NULL if there's nothing to do at request end */
	PHP_MINFO(mcache),
#if ZEND_MODULE_API_NO >= 20010901
	"0.1", /* Replace with version number for your extension */
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

ZEND_DECLARE_MODULE_GLOBALS(mcache);

#ifdef COMPILE_DL_MCACHE
ZEND_GET_MODULE(mcache)
#endif

static void mc_resource_list_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC) /* {{{ */
{
  struct memcache_wrapper *mc = (struct memcache_wrapper *)rsrc->ptr;
  mcm_free(mc->ctxt, mc->mc);
  efree(mc);
}
/* }}} */

static void mc_persistent_resource_list_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC) /* {{{ */
{
  struct memcache_wrapper *mc = (struct memcache_wrapper *)rsrc->ptr;
  mcm_free(mc->ctxt, mc->mc);
  free(mc);
  MCACHE_G(num_persistent)--;
}
/* }}} */

static void php_mcache_init_globals(zend_mcache_globals *mcache_globals_p TSRMLS_DC) /* {{{ */
{
	MCACHE_G(num_persistent) = 0;
}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(mcache)
{
	zend_class_entry mcache_class_entry;

  persist_ctx = mcMemNewCtxt(free, malloc, malloc, realloc);
  standard_ctx = mcMemNewCtxt(free, malloc, malloc, realloc);
  mcMemSetupCtxt(persist_ctx, xpefree, xpemalloc, xpemalloc, xperealloc);
  mcMemSetupCtxt(standard_ctx, xpefree, xpemalloc, xpemalloc, xperealloc);
    
  
	mcErrSetupCtxt(persist_ctx, (mcErrFunc)mcache_err);
	mcErrSetupCtxt(standard_ctx, (mcErrFunc)mcache_err);
	mcMemSetup(xefree, (mcMallocFunc)xemalloc, (mcMallocFunc)xemalloc, (mcReallocFunc)xerealloc);
	mcErrSetupCtxt(standard_ctx, (mcErrFunc)mcache_err);
	le_mcache = zend_register_list_destructors_ex(mc_resource_list_dtor, NULL, "memcache object", module_number);
	le_pmcache = zend_register_list_destructors_ex(NULL, mc_persistent_resource_list_dtor, "persistent memcache object", module_number);
	
  INIT_CLASS_ENTRY(mcache_class_entry, "Memcache", mcache_class_functions);
  mcache_class_entry_ptr = zend_register_internal_class(&mcache_class_entry TSRMLS_CC);

#ifdef ZTS
	ts_allocate_id(&mcache_globals_id, sizeof(zend_mcache_globals), (ts_allocate_ctor) php_mcache_init_globals, NULL);
#else
	php_mcache_init_globals(&mcache_globals TSRMLS_CC);
#endif

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(mcache)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(mcache)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(mcache)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(mcache)
{
	char buf[16];

	snprintf(buf, 16, "%ld", MCACHE_G(num_persistent));
	
	php_info_print_table_start();
	php_info_print_table_header(2, "mcache support", "enabled");
	php_info_print_table_row(2, "open persistent connections", buf);
	php_info_print_table_row(2, "linked libmemcache version", mc_version());
	php_info_print_table_row(2, "mcache extension version", "1.2.0-beta10 (compatible w/libmemcache 1.4.0b9)");
	php_info_print_table_end();
}
/* }}} */

/**
 * Internal usage functions
 */

static int mc_get_resource(zval *obj, struct memcache_wrapper **mc TSRMLS_DC) /* {{{ */
{
    zval    **resource;
    int     resource_type;

    if (Z_TYPE_P(obj) != IS_OBJECT || zend_hash_find(Z_OBJPROP_P(obj), "resource_id", sizeof("resource_id"), (void **)&resource) == FAILURE) {
        zend_error(E_WARNING, "Cannot find memcache object/resources");
        return 0;
    }

    *mc = (struct memcache_wrapper *) zend_list_find(Z_LVAL_PP(resource), &resource_type);

    if (!*mc || (resource_type != le_mcache && resource_type != le_pmcache )) {
        zend_error(E_WARNING, "Invalid memcache object/resource found");
        return 0;
    }

    return Z_LVAL_PP(resource);
}
/* }}} */

static int mc_get_storage_data(INTERNAL_FUNCTION_PARAMETERS, smart_str *buf, char **key, int *key_len, char **data, int *data_len, long *flags, long *expire) /* {{{ */ {
	zval *val;

	php_serialize_data_t val_hash;

	*data = NULL;
	*data_len = 0;
	*key = NULL;
	*key_len = 0;
	*flags = 0;
	*expire = 0;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sz|l",
				key, key_len, &val, expire)) {
		return 1;
	}

	switch (Z_TYPE_P(val)) {
		case IS_LONG:
			convert_to_string(val);
			*flags = MC_LONG_VAL;
			*data = Z_STRVAL_P(val);
			*data_len = Z_STRLEN_P(val);
			break;
		case IS_DOUBLE:
			convert_to_string(val);
			*flags = MC_DOUBLE_VAL;
			*data = Z_STRVAL_P(val);
			*data_len = Z_STRLEN_P(val);
			break;
		case IS_STRING:
			*data = Z_STRVAL_P(val);
			*data_len = Z_STRLEN_P(val);
			break;
		default:
			PHP_VAR_SERIALIZE_INIT(val_hash);
			php_var_serialize(buf, &val, &val_hash TSRMLS_CC);
			PHP_VAR_SERIALIZE_DESTROY(val_hash);

			*flags = MC_SERIALIZED;

			*data = buf->c;
			*data_len = buf->len;
			break;
	}
	return 0;

}
/* }}} */

/* }}} */

/**
 * PHP exposed functions
 */

/* {{{ proto object pmemcache(string identifier, [int expiration])
   Creates/Retrieves persisten named Memcache object */
PHP_FUNCTION(pmemcache) {
	list_entry new_le, *le;
	int id;
	char *name, *hash_key;
	int name_len, hash_key_len;
	long expiration = 0;
	struct memcache_wrapper *mc = NULL;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l",
				&name, &name_len, &expiration)) {
		return;
	}

	hash_key_len = strlen("mcache_pconnect___:")+name_len;
	hash_key = emalloc(hash_key_len+1);
	snprintf(hash_key, hash_key_len+1, "mcache_pconnect___:%s", name);
	
	if(zend_hash_find(&EG(persistent_list), hash_key, hash_key_len+1, (void **)&le) == SUCCESS) {
		if(le->type != le_pmcache) {
			zend_error(E_WARNING, "Invalid persistent Memcache object.");
			efree(hash_key);
			RETURN_FALSE;
		}

		mc = le->ptr;
		
		// check that the object isn't expired, if it is clean it up and remove from list, then allocate
		// a new object for this request. (0 signifies never expire)
		if(mc->expiration != 0) {
			time_t now = time(NULL);
			if(now > mc->expiration) {
				if(zend_hash_del(&EG(persistent_list), hash_key, hash_key_len+1) != SUCCESS) {
					zend_error(E_WARNING, "Failed deleting expired persistent memcache object.");
					efree(hash_key);
					RETURN_FALSE;
				}
				goto new_object;	
			}
		}
		
	} else {
new_object:
		mc = malloc(sizeof(struct memcache_wrapper));
		mc->is_initialized = 0;
		if(expiration) {
			mc->expiration = time(NULL)+expiration;
		} else {
			mc->expiration = 0;
		}
		mc->mc = mcm_new(persist_ctx);
		mc->ctxt = persist_ctx;
		if(mc->mc == NULL) {
			zend_error(E_WARNING, "Can't allocate new memcache object.");
			efree(hash_key);
			free(mc);
			RETURN_FALSE;
		}

		new_le.type = le_pmcache;
		new_le.ptr = mc;
		if(zend_hash_update(&EG(persistent_list), hash_key, hash_key_len+1, (void *)&new_le,
				sizeof(list_entry), NULL)==FAILURE) {
			zend_error(E_NOTICE, "Unable to add persistent Memcache object to list.");
		}
		MCACHE_G(num_persistent)++;
	}
	
	if(mc->mc) {
		id = zend_list_insert(mc, le_pmcache);
		object_init_ex(return_value, mcache_class_entry_ptr);
		add_property_resource(return_value, "resource_id", id);
		efree(hash_key);
		return;
	}

	if(mc) {
		free(mc);
	}
	efree(hash_key);
	RETURN_FALSE;
}
/* }}} */

/* {{{ proto object memcache()
   Creates new Memcache object */
PHP_FUNCTION(memcache) {
	list_entry new_le;
	int id;

	struct memcache_wrapper *mc;

	if(ZEND_NUM_ARGS()!=0) {
		WRONG_PARAM_COUNT;
	}

	mc = emalloc(sizeof(struct memcache_wrapper));
	mc->is_initialized = 0;
	mc->expiration = 0;
	mc->mc = mcm_new(standard_ctx);
	mc->ctxt = standard_ctx;
	if(mc->mc == NULL) {
		zend_error(E_WARNING, "Can't allocate new memcache object.");
		efree(mc);
		RETURN_FALSE;
	}

	id = zend_list_insert(mc, le_mcache);
	object_init_ex(return_value, mcache_class_entry_ptr);
	add_property_resource(return_value, "resource_id", id);

	return;
}
/* }}} */

/* {{{ proto bool is_initialized()
   Tells you whether this memcache object is initialized, if so you want to do add_server, otherwise you don't. */
PHP_FUNCTION(is_initialized) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_FALSE;
	}

	if(mc->is_initialized) {
		RETURN_TRUE;
	}

	RETURN_FALSE;
}
/* }}} */

/* {{{ proto bool set_initialized()
   Sets the memcache object as initialized/unitialized. */
PHP_FUNCTION(set_initialized) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;
	zend_bool init_bool = 1;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|b",
				&init_bool)) {
		return;
	}
	
	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_FALSE;
	}

	if(init_bool) {
		mc->is_initialized = 1;
	} else {
		mc->is_initialized = 0;
	}

	RETURN_TRUE;
}
/* }}} */

/* {{{ proto bool add_server( string host , string port )
   Connects to memcache server */
PHP_FUNCTION(add_server) {
	zval *mc_obj = getThis();
	char *host, *port;
	int host_len, port_len;
	struct memcache_wrapper *mc;
	int resource;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss",
				&host, &host_len, &port, &port_len)) {
		return;
	}

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_FALSE;
	}

	mcm_server_add2(mc->ctxt, mc->mc, host, host_len, port, port_len);
	
	RETURN_TRUE;
}
/* }}} */

/* {{{ proto mixed get( string key )
   Attempts to get a key from the server, returns false on miss or failure. */
PHP_FUNCTION(get) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;
	zval *arg;
	
	struct memcache_req *req;
	struct memcache_res *res;
	
	char *key;
	int key_len;

	php_unserialize_data_t val_hash;

	int is_multi = 0;
	
	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z",
				&arg)) {
		return;
	}

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_NULL();
	}

	req = mcm_req_new(mc->ctxt);
	if(req == NULL) {
		zend_error(E_ERROR, "Could not allocate new mc_req");
		RETURN_NULL();
	}

	switch (Z_TYPE_P(arg)) {
		case IS_STRING:
			convert_to_string(arg);
			key = Z_STRVAL_P(arg);
			key_len = Z_STRLEN_P(arg);
			break;
		case IS_ARRAY:
			is_multi = 1;
			break;
		default:
			zend_error(E_ERROR, "Invalid arguments to memcache->get()");
			mcm_req_free(mc->ctxt, req);
			RETURN_NULL();
	}
									
	if(!is_multi) {
		res = mcm_req_add_ref(mc->ctxt, req, key, key_len);

		if(res == NULL) {
			zend_error(E_ERROR, "Could not allocate new mc_res");
			mcm_req_free(mc->ctxt, req);
			RETURN_NULL();
		}

		mcm_get(mc->ctxt, mc->mc, req);

		if(mcm_res_found(mc->ctxt, res) && res->bytes) {
			if(res->flags & MC_SERIALIZED) {
				const char *temp_str = (const char *)res->val;
				
				PHP_VAR_UNSERIALIZE_INIT(val_hash);
				if (!php_var_unserialize(&return_value, (const unsigned char **) &temp_str, res->val + res->bytes, &val_hash TSRMLS_CC)) {
					PHP_VAR_UNSERIALIZE_DESTROY(val_hash);
					zend_error(E_WARNING, "Error unserializing data.");
					mcm_req_free(mc->ctxt, req);
					RETURN_NULL();
				} else {
					PHP_VAR_UNSERIALIZE_DESTROY(val_hash);
					mcm_req_free(mc->ctxt, req);
					return;
				}
			} else {
				ZVAL_STRINGL(return_value, res->val, res->bytes, 1);

				if(res->flags & MC_DOUBLE_VAL) {
					convert_to_double(return_value);
				} else if(res->flags & MC_LONG_VAL) {
					convert_to_long(return_value);
				}
				
				mcm_req_free(mc->ctxt, req);
				return;
			}
		} else {
			mcm_req_free(mc->ctxt, req);
			RETURN_NULL();
		}

	} else {
		int cnt;
		zval **data, *tmp;
		struct memcache_res **res_pp, *cur_res;
		HashTable *htable;
		int i = 0;
		char *hkey;
		ulong index;
		
		array_init(return_value);
		cnt = zend_hash_num_elements(Z_ARRVAL_P(arg));
		
		// Allocate enough room for each request to go in an array,
		// this because iterating through the req object sucks!
		res_pp = emalloc(sizeof(struct memcache_res)*cnt);
		
		htable = HASH_OF(arg);
		zend_hash_internal_pointer_reset(htable);
		while(zend_hash_get_current_key(htable, &key, &index, 0) == HASH_KEY_IS_LONG) {
			zend_hash_get_current_data(htable, (void **)&data);
			convert_to_string(*data);
			key = Z_STRVAL_P(*data);
			key_len = Z_STRLEN_P(*data);

			res_pp[i] = mcm_req_add_ref(mc->ctxt, req, key, key_len);

			if(res_pp[i] == NULL) {
				zend_error(E_ERROR, "Could not allocate new memcache_res_pp");
				mcm_req_free(mc->ctxt, req);
				return;
			}
			zend_hash_move_forward(htable);
			++i;
		}

		mcm_get(mc->ctxt, mc->mc, req);

		for(i=0; i<cnt; ++i) {
			cur_res = res_pp[i];
			if(mcm_res_found(mc->ctxt, cur_res) && cur_res->bytes) {
				MAKE_STD_ZVAL(tmp);
				if(cur_res->flags & MC_SERIALIZED) {
					const char *temp_str = (const char *)cur_res->val;
				
					PHP_VAR_UNSERIALIZE_INIT(val_hash);
					if (!php_var_unserialize(&tmp, (const unsigned char **) &temp_str, cur_res->val + cur_res->bytes, &val_hash TSRMLS_CC)) {
						PHP_VAR_UNSERIALIZE_DESTROY(val_hash);
						zend_error(E_WARNING, "Error unserializing data.");
						zval_dtor(tmp);
						efree(tmp);
					} else {
						PHP_VAR_UNSERIALIZE_DESTROY(val_hash);
						add_assoc_zval(return_value, (char *)cur_res->key, tmp);
					}
				} else {
					ZVAL_STRINGL(tmp, cur_res->val, cur_res->bytes, 1);
					
					if(cur_res->flags & MC_DOUBLE_VAL) {
						convert_to_double(tmp);
					} else if(cur_res->flags & MC_LONG_VAL) {
						convert_to_long(tmp);
					}
					
					add_assoc_zval(return_value, (char *)cur_res->key, tmp);
				}
			}
		}

		mcm_req_free(mc->ctxt, req);
		efree(res_pp);
		return;
	}
}
/* }}} */

/* {{{ proto bool set( string key, mixed val, [ int expire ])
   Sets an item in the cache to val. */
PHP_FUNCTION(set) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;
	smart_str buf = {0};
	
	char *key, *data;
	int key_len, data_len;
	long flags, expire;
	int ret;

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_FALSE;
	}

	if(mc_get_storage_data(INTERNAL_FUNCTION_PARAM_PASSTHRU, &buf, &key, &key_len, &data, &data_len, &flags, &expire)!=0) {
		RETURN_FALSE;
	}

	ret=mcm_set(mc->ctxt, mc->mc, key, key_len, data, data_len, expire, flags);
	smart_str_free(&buf);
	
	if(ret==0) {
		RETURN_TRUE;
	}
	
	RETURN_FALSE;
}
/* }}} */

/* {{{ proto bool add( string key, mixed val, [ int expire ])
   Adds an item in the cache if it doesn't already exist. */
PHP_FUNCTION(add) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;
	smart_str buf = {0};
	
	char *key, *data;
	int key_len, data_len;
	long flags, expire; 
	int ret;
	
	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_FALSE;
	}

	if(mc_get_storage_data(INTERNAL_FUNCTION_PARAM_PASSTHRU, &buf, &key, &key_len, &data, &data_len, &flags, &expire)!=0) {
		RETURN_FALSE;
	}

	ret=mcm_add(mc->ctxt, mc->mc, key, key_len, data, data_len, expire, flags);
	smart_str_free(&buf);
	
	if(ret==0) {
		RETURN_TRUE;
	}
	
	RETURN_FALSE;
}
/* }}} */

/* {{{ proto bool replace( string key, mixed val, [ int expire ])
   Replaces an item in the cache if it already exists. */
PHP_FUNCTION(replace) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;
	smart_str buf = {0};
	
	char *key, *data;
	int key_len, data_len;
	long flags, expire; 
	int ret;
	
	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_FALSE;
	}

	if(mc_get_storage_data(INTERNAL_FUNCTION_PARAM_PASSTHRU, &buf, &key, &key_len, &data, &data_len, &flags, &expire)!=0) {
		RETURN_FALSE;
	}

	ret=mcm_replace(mc->ctxt, mc->mc, key, key_len, data, data_len, expire, flags);
	smart_str_free(&buf);
	
	if(ret==0) {
		RETURN_TRUE;
	}
	
	RETURN_FALSE;
}
/* }}} */

/* {{{ proto long incr( string key, [int value])
   Sets an item in the cache to val. */
PHP_FUNCTION(incr) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;

	char *key;
	int key_len;
	long value = 1;
	int ret;
	
	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l",
				&key, &key_len, &value)) {
		return;
	}

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_NULL();
	}

	ret=mcm_incr(mc->ctxt, mc->mc, key, key_len, value);

	if(ret == 0 && mc->ctxt->errnum == ENOENT) {
		RETURN_NULL();
	}
	
	RETURN_LONG(ret);
}
/* }}} */

/* {{{ proto long decr( string key, [int value])
   Sets an item in the cache to val. */
PHP_FUNCTION(decr) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;

	char *key;
	int key_len;
	long value = 1;
	int ret;
	
	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l",
				&key, &key_len, &value)) {
		return;
	}

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_NULL();
	}

	ret=mcm_decr(mc->ctxt, mc->mc, key, key_len, value);
	
	if(ret == 0 && mc->ctxt->errnum == ENOENT) {
		RETURN_NULL();
	}
	
	RETURN_LONG(ret);
}
/* }}} */

/* {{{ proto bool delete( string key, [ int expire ])
   Deletes existing item */
PHP_FUNCTION(delete) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;

	char *key;
	int key_len;
	long hold_time = 0;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l",
			&key, &key_len, &hold_time)) {
		return;
	}

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_FALSE;
	}
	
	
	if(mcm_delete(mc->ctxt, mc->mc, key, key_len, (const time_t)hold_time)==0) {
		RETURN_TRUE;
	}
	
	RETURN_FALSE;
}
/* }}} */

/* {{{ proto bool flush_all()
 * Flushes all keys from all connected servers.*/
PHP_FUNCTION(flush_all) {
	zval *mc_obj = getThis();
	struct memcache_wrapper *mc;
	int resource;

	if(ZEND_NUM_ARGS()!=0) {
		WRONG_PARAM_COUNT;
	}

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_FALSE;
	}

	if(mcm_flush_all(mc->ctxt, mc->mc)==0) {
		RETURN_TRUE;
	} else {
		RETURN_FALSE;
	}
}
/* }}} */

/* {{{ proto array stats()
   Returns associative array with various server statistics. */
PHP_FUNCTION(stats) {
	zval *mc_obj = getThis();
	
	struct memcache_server_stats *s;
	struct memcache_wrapper *mc;
	int resource;

	if((resource = mc_get_resource(mc_obj, &mc TSRMLS_CC))==0) {
		RETURN_NULL();
	}
	
	s = mcm_stats(mc->ctxt, mc->mc);
	if(s==NULL) {
		RETURN_NULL();
	}
	
	array_init(return_value);
	add_assoc_long(return_value, "pid", s->pid);
	add_assoc_long(return_value, "uptime", s->uptime);
	add_assoc_long(return_value, "time", s->time);
	if(s->version) {
		add_assoc_string(return_value, "version", s->version, 1);
	} else {
		add_assoc_string(return_value, "version", "Unknown", 1);
	}
	add_assoc_double(return_value, "usage_user", s->rusage_user.tv_sec+s->rusage_user.tv_usec/1000000);
	add_assoc_double(return_value, "usage_system", s->rusage_system.tv_sec+s->rusage_system.tv_usec/1000000);
	add_assoc_double(return_value, "curr_items", s->curr_items);
	add_assoc_double(return_value, "total_items", s->total_items);
	add_assoc_double(return_value, "bytes", s->bytes);
	add_assoc_double(return_value, "curr_connections", s->curr_connections);
	add_assoc_double(return_value, "total_connections", s->total_connections);
	add_assoc_double(return_value, "connection_structures", s->connection_structures);
	add_assoc_double(return_value, "cmd_get", s->cmd_get);
	add_assoc_double(return_value, "cmd_set", s->cmd_set);
	add_assoc_double(return_value, "get_hits", s->get_hits);
	add_assoc_double(return_value, "get_misses", s->get_misses);
	add_assoc_double(return_value, "bytes_read", s->bytes_read);
	add_assoc_double(return_value, "bytes_written", s->bytes_written);
	add_assoc_double(return_value, "limit_maxbytes", s->limit_maxbytes);

	mcm_server_stats_free(mc->ctxt, s);
	
	return;	
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
